#!/usr/bin/python
import copy

import gcc_eval as gcc

def move_idx_to_delta(idx):
    if idx == 0: return (0, -1)
    if idx == 1: return (1, 0)
    if idx == 2: return (0, 1)
    if idx == 3: return (-1, 0)
    return None

class Game(object):
    def __init__(self, field_str, lman_prog):
        self._parse_field(field_str)
        self.lman_ai = gcc.GccMachine(lman_prog, [], [], None)
        self.score = 0

    def _parse_field(self, s):
        self.ghost_coords = []
        self.map = []
        for y, l in enumerate(s.splitlines()):
            line = list(l)
            for x, c in enumerate(line):
                if c == '\\':
                   self.lman_coords = [x, y]
                elif c == '%':
                    self.fruit_coords = [x, y]
                elif c == '=':
                    self.ghost_coords.append([x, y])
            self.map.append(line)
        self.map_dim = (len(self.map[0]), len(self.map))

        self.ghost_coords = [(16,9)]

    def _lman_prepare_world_state(self):
        char_to_code = {'#': 0, ' ': 1, '.': 2, 'o': 3, '%':4, '\\': 5, '=': 6}

        w_map = []
        for line in self.map:
            enc_line = [char_to_code[c] for c in line]
            w_map.append(gcc.to_gcc_list(enc_line))
        w_map = gcc.to_gcc_list(w_map)

        lman_status = gcc.to_gcc_tuple([
            0, # vitality
            tuple(self.lman_coords),
            0, # direction -- FIXME
            3, # number of lives -- FIXME
            self.score
        ])

        ghost_status = []
        for x, y in self.ghost_coords:
            ghost_status.append(gcc.to_gcc_tuple((1, gcc.to_gcc_tuple((x, y)), 'dir')))
        ghost_status = gcc.to_gcc_list(ghost_status)

        fruit_status = []

        return gcc.to_gcc_tuple((w_map, lman_status, ghost_status, fruit_status))

    def _move_lman(self, move_idx):
        delta = move_idx_to_delta(move_idx)
        print 'MOVE', move_idx
        if not delta:
            print 'ILLEGAL MOVE', move_idx
            return

        coords = [c + d for c, d in zip(self.lman_coords, delta)]
        out_of_range = [(c < 0) and (c > lim)
                            for c, lim in zip(coords, self.map_dim)]
        if any(out_of_range):
            print 'Move outside map borders'
            return

        x, y = coords
        c = self.map[y][x]
        if c == '#':
            return

        if c == '.':
            self.score += 10
            self.map[y][x] = ' '

        if c == 'o':
            self.score += 50
            self.map[y][x] = ' '

        self.lman_coords = coords
        print 'moved lman to', coords

    def lman_start(self):
        self.lman_ai.d = [(None, 'STOP')]
        self.lman_ai.s = []
        self.lman_ai.e = gcc.Frame(None,
                    [self._lman_prepare_world_state(), 'UNDEFINED'])
        self.lman_ai.c = 0

        self.lman_ai.run()
        self.ai_state, self.ai_step_f = self.lman_ai.s.pop()

    def lman_step(self, debug=1):
        self.lman_ai.d = [(None, 'STOP')]
        self.lman_ai.s = []
        self.lman_ai.e = gcc.Frame(None,
                    [self.ai_state, self._lman_prepare_world_state()])
        self.lman_ai.c = self.ai_step_f.fp

        n_ops = self.lman_ai.run(debug=debug)
        self.ai_state, move = self.lman_ai.s.pop()
        self._move_lman(move)
        return n_ops

    def print_map(self):
        objs = {}
        objs[tuple(self.lman_coords)] = '\\'
        for gc in self.ghost_coords:
            objs[tuple(gc)] = 'W'

        def subst(c, x, y):
            obj = objs.get((x,y))
            if obj != None:
                return obj
            return c if c in ' #.o' else ' '

        for y, line in enumerate(self.map):
            flt = [subst(c, x, y) for x, c in enumerate(line)]
            print ''.join(flt)


if __name__ == '__main__':
    import sys
    map_data = open(sys.argv[1]).read()
    lman_prog = open(sys.argv[2]).read()
    g = Game(map_data, lman_prog)
    g.lman_start()

    debug = 1
    while 1:
        n_ops = g.lman_step(debug=debug)
        g.print_map()
        print '--- score %d, step done in %d ops, press <enter> ' \
              'or enter new debuglevel' % (g.score, n_ops)
        inp = raw_input()
        inp = inp.strip()
        if inp:
            debug = int(inp)
