#!/usr/bin/env python
# install python-funcparserlib

from pprint import pformat, pprint
import re
from StringIO import StringIO
import sys

from tokenize import generate_tokens
import token
from funcparserlib.parser import (skip, many, some, maybe,
                                   with_forward_decls, finished)


class Token(object):
    def __init__(self, code, value, start=(0, 0), stop=(0,0), line=''):
        self.code = code
        self.value = value
        self.start = start
        self.stop = stop
        self.line = line

    @property
    def type(self):
        return token.tok_name[self.code]

    def __unicode__(self):
        pos = '-'.join('%d,%d' % x for x in [self.start, self.stop])
        return "%s %s '%s'" % (pos, self.type, self.value)

    def __repr__(self):
        return 'Token(%r, %r, %r, %r, %r)' % (
                self.code, self.value, self.start, self.stop, self.line)

    def __eq__(self, other):
        return (self.code, self.value) == (other.code, other.value)


def tokenize(s):
    return [Token(*t) for t in generate_tokens(StringIO(s).readline)]

def token_value(t):
    return t.value

def flatten_expr(arg):
    # flatten:
    # (Variable(a), Operator(-), (6, [(Operator(+), 5), (Operator(+), Variable(v))]))
    return [arg[0]] + [i for pair in arg[1] for i in pair]

def fix_varlist(arg):
    return [arg[0]] + [pair[1] for pair in arg[1]]

def unpack(f):
    return (lambda args: f(*args) if isinstance(args, tuple) else f(args))

class Nop(object):
    def __init__(self):
        pass

    def __repr__(self):
        return 'Nop'

class Asm (object):
    def __init__(self,opcode,params):
        self.command=opcode+" "+"".join(map(lambda x:(x.name.replace("%","")) if (x.__class__ in [Operator,Variable])else `x`,params))
    def __repr__(self):
        return "asm " +self.command

class Variable(object):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'Variable(%s)' % self.name


class Operator(object):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'Operator(%s)' % self.name


class OperatorAssignment(object):
    def __init__(self, left, atom1, op, atom2):
        self.left = left
        self.atom1 = atom1
        self.op = op
        self.atom2 = atom2

    def __repr__(self):
        return '%s = %s %s %s' % (self.left, self.atom1, self.op, self.atom2)



class SimpleAssignment(object):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __repr__(self):
        return '%s = %s' % (self.left, self.right)


class Return(object):
    def __init__(self, resultlist):
        self.resultlist = resultlist

    def __repr__(self):
        return 'return %s' % (self.resultlist,)

class ReturnFun(object):
    def __init__(self, name,arglist):
        self.arglist = arglist
        self.name=name
    def __repr__(self):
        return 'returnfun %s(%s)' % (self.name,self.arglist,)


class FunctionCall(object):
    def __init__(self, resultlist, name, arglist):
        self.resultlist = resultlist
        self.name = name
        self.arglist = arglist

    def __repr__(self):
        return '%s = %s(%s)' % (self.resultlist, self.name, self.arglist)


class If:
    def __init__(self,left,op,right,truelist,falselist):
        self.left=left
        self.op=op
        self.right=right
        self.truelist=truelist
        self.falselist=falselist

    def __repr__(self):
        truelines = ['   %s  %s' % (i.__class__.__name__,i) for i in self.truelist]
        falselines = ['   %s  %s' % (i.__class__.__name__,i) for i in self.falselist]
        return 'If (%s %s %s) {\n%s } else {\n%s} ' % (
                    self.left, self.op, self.right,
                    "\n".join(truelines),"\n".join(falselines))


class While(object):
    def __init__(self,cond,exprlist):
        self.cond=cond
        self.exprlist=exprlist

    def __repr__(self):
        lines = ['   %s  %s' % (i.__class__.__name__,i) for i in self.exprlist]
        return 'while (%s) {\n%s }'%(self.cond,"\n".join(lines))


class FunctionDef(object):
    def __init__(self, name, arglist, locallist, exprlist):
        self.name = name
        self.arglist = arglist
        self.locallist= locallist
        self.exprlist = exprlist

    def __repr__(self):
        lines = ['   %s  %s' % (i.__class__.__name__,i) for i in self.exprlist]
        return 'def %s(%s:%s) {\n%s\n}\n' % (self.name, self.arglist,self.locallist, '\n'.join(lines))


class Program(object):
    def __init__(self, funclist):
        self.funclist = funclist

    def __repr__(self):
        return '\n'.join(str(f) for f in self.funclist)


def varno(varlist,var,fno):
    return varlist.index(var.name)+fno*32

class Assemble(object):
    def __init__(self,program):
        self.funcref={}#name ->
        self.funbody={}
        self.funifs={}
        self.ifno=1
        self.whileno=1
        self.jumpno=1
        self.funclist=map(lambda x:x.name,program.funclist)
        print "####################3333%s"%self.funclist
        for fun in program.funclist:
            print "funcref "+fun.name+" "+`(len(fun.arglist),len(fun.locallist))`
            self.funcref[fun.name]=(len(fun.arglist),len(fun.locallist))
        
        for fun in enumerate(program.funclist):
            if fun[1].__class__ ==FunctionDef:
                varlist=map(lambda x:x.name,fun[1].arglist)+map(lambda x:x.name,fun[1].locallist)
                (self.funbody[fun[1].name],self.funifs[fun[1].name])= self.gen_asm(varlist,fun[1].exprlist,fun[0])
        self.out=[]
        self.funaddr={}
        #print self.funbody
        self.out.extend(self.funbody['main'])
        
        for fun in self.funbody.keys():
            if fun<>'main':
                self.funaddr[fun]=len(self.out)
                self.out.extend(self.funbody[fun])
               
        for line in range(len(self.out)):
            mo=re.search("j[ltgeq]+ %if(\d+),.+",self.out[line])
            if mo:
                for j in range(line,len(self.out)):
                    if re.search(("%%if%strue"%mo.group(1)),self.out[j]):
                        self.out[line]=(self.out[line]).replace("%if"+mo.group(1),`j`)
                        break

            mo=re.search("j[ltgeq]+ %if(\d+)afterend.+",self.out[line])
            if mo:
                for j in range(line,len(self.out)):
                    if re.search(("%%if%send"%mo.group(1)),self.out[j]):
                        self.out[line]=(self.out[line]).replace("%if"+mo.group(1)+"afterend",`j+1`)
                        break

            mo=re.search("j[ltgeq]+ %while(\d+),.+",self.out[line])
            if mo:
                for j in range(line,len(self.out)):
                    if re.search(("%%while%sstart"%mo.group(1)),self.out[j]):
                        self.out[line]=(self.out[line]).replace("%while"+mo.group(1),`j`)
                        break

            mo=re.search("j[ltgeq]+ %while(\d+)afterend.+",self.out[line])
            if mo:
                for j in range(line,len(self.out)):
                    if re.search(("%%while%send"%mo.group(1)),self.out[j]):
                        self.out[line]=(self.out[line]).replace("%while"+mo.group(1)+"afterend",`j+1`)
                        self.out[j]=(self.out[j]).replace("%while"+mo.group(1)+"end",`line-1`,1)
                        
                        break
        for line in range(len(self.out)):
            self.out[line]=re.sub("#[a-zA-Z0-9]+",lambda x:`self.funaddr[x.group(0)[1:]]`,self.out[line])

    def __repr__(self):

        return "\n".join(self.out)

    def gen_asm(self,varlist,exlist,fno):
        out =[]
        ifs=[]
        for ex in exlist:
            print ex
            if ex.__class__==SimpleAssignment:
                if ex.left.__class__==Variable:
                    if ex.right.__class__==Variable:
                        out.append('mov [%s],[%s]'%(varno(varlist,ex.left,fno),varno(varlist,ex.right,fno)))
                    else:
                        out.append('mov [%s],%s'%(varno(varlist,ex.left,fno),ex.right))
                else:
                    raise Exception("illegal SimpleAssignment" + ex)

            elif ex.__class__==OperatorAssignment:
                if (ex.atom1.__class__==int) or (ex.atom1.name!=ex.left.name):
                    out.append('mov [%s],[%s]'%(varno(varlist,ex.left,fno),varno(varlist,ex.atom1,fno)))

                if ex.op.name=="+":
                    opcode="ADD"
                elif ex.op.name=="-":
                    opcode="SUB"
                elif ex.op.name=="*":
                    opcode="MUL"
                elif ex.op.name=="/":
                    opcode="DIV"
                elif ex.op.name=="^":
                    opcode="XOR"
                elif ex.op.name=="|":
                    opcode="OR"
                elif ex.op.name=="&":
                    opcode="AND"
                else:
                    raise Exception ("Illegal Operator %s"%ex.op)
                if ex.atom2.__class__==Variable:
                    out.append("%s [%s],[%s]"%(opcode,varno(varlist,ex.left,fno),varno(varlist,ex.atom2,fno)))
                else:
                    out.append('%s [%s],%s'%(opcode,varno(varlist,ex.left,fno),ex.atom2))
            elif ex.__class__==FunctionCall:
                fref=self.funcref[ex.name]
                if fref[0]!=len(ex.arglist):
                    raise Exception ("wrong arg count %s",ex)
                left=varno(varlist,ex.resultlist[0],fno)
                for ar in enumerate(ex.arglist):
                    out.append('mov [%s],[%s]'%(self.funclist.index(ex.name)*32+ar[0],varno(varlist,ar[1],fno)))
                out.append('mov [%s],pc'%(self.funclist.index(ex.name)*32+31))
                out.append('jeq #%s,0,0'%ex.name)
                out.append('mov [%s],[%s]'%(left,self.funclist.index(ex.name)*32+30)) 
            elif ex.__class__==Return:
                ret=ex.resultlist[0]
                if ret.__class__==Variable:
                    out.append("mov [%s],[%s]"%(fno*32+30,varno(varlist,ret,fno)))
                else:
                    out.append("mov [%s],[%s]"%(fno*32+30,ret))
                    
                out.append("add [%s],2"%(fno*32+31))
                #out.append("jeq [%s],0,0"%(fno*32+31))
                out.append("mov pc,[%s]"%(fno*32+31))
            elif ex.__class__==If:
                curif=self.ifno
                self.ifno+=1
                if ex.op.name in [">"]:
                    opcode='jgt'
                if ex.op.name in ["<"]:
                    opcode='jlt'
                if ex.op.name in ["=="]:
                    opcode='jeq'
                
                if ex.left.__class__==Variable:
                    left='[%s]'%varno(varlist,ex.left,fno)
                else:
                    left=ex.left
                if ex.right.__class__==Variable:
                    right='[%s]'%varno(varlist,ex.right,fno)
                else:
                    right=ex.right
                    
                out.append("%s %%if%s,%s,%s;if%sstart"%(opcode,curif,left,right,curif))
                
                truecode,trueifs=self.gen_asm(varlist,ex.truelist,fno)
                truecode[0]+=';%%if%strue'%curif
                truecode[-1]+=';%%if%send'%curif
                falsecode,falseifs=self.gen_asm(varlist,ex.falselist,fno)
                out.extend(falsecode)
                out.append('jeq %%if%safterend,0,0;if%sfalseend'%(curif,curif))
                out.extend(truecode)
                
            elif ex.__class__==Asm:
#                if ex.command.split()[-1]
                out.append(ex.command)
            elif ex.__class__==Nop:
                pass
            elif ex.__class__==While:
                curwhile=self.whileno
                self.whileno+=1
                if ex.cond.__class__==Variable:
                    cond='[%s]'%varno(varlist,ex.cond,fno)
                else:
                    cond=ex.cond
                
                out.append("jgt %%while%s,%s,0 ;while%s"%(curwhile,cond,curwhile))
                out.append('jeq %%while%safterend,0,0'%(curwhile))
                whilecode,whileifs=self.gen_asm(varlist,ex.exprlist,fno)
                whilecode[0]+=";%%while%sstart"%curwhile
                out.extend(whilecode)
                out.append("jeq %%while%send,0,0;%%while%send"%(curwhile,curwhile))

            else:
                raise Exception("Illegal expression %s"% ex)


        return (out,ifs)



def mk_op(name):
    return (some(lambda t: t.type == 'OP' and t.value == name) >>
            token_value >>
            (lambda _: Operator(name)))

def mk_keyword(name):
    return (some(lambda t: t.type == 'NAME' and t.value == name) >>
            token_value)

#empty_line = some(lambda t: t.type == 'NL') >> (lambda x: Nop())
newline = some(lambda t: t.type in ('NEWLINE', 'NL')) >> (lambda x: Nop())

number = some(lambda t: t.type == 'NUMBER') >> token_value >> int
function_name = some(lambda t: t.type == 'NAME') >> token_value
variable = some(lambda t: t.type == 'NAME') >> token_value >> Variable
simple_expr = number | variable
variable_list = variable + many(mk_op(',') + variable) >> fix_varlist
local_list = variable + many(mk_op(',') + variable) >> fix_varlist

simple_expr_list = simple_expr + many(mk_op(',') + simple_expr) >> fix_varlist
arith_op = mk_op('*') | mk_op('/') | mk_op('+') | mk_op('-') |mk_op('|')|mk_op('&')|mk_op('^')
comp_op = mk_op('<') | mk_op('>') | mk_op('==') | mk_op('>=') | mk_op('<=')

lparen = skip(mk_op('('))
rparen = skip(mk_op(')'))
endmarker = some(lambda t: t.type == 'ENDMARKER')

operator_assignment = (
    variable + skip(mk_op('=')) + simple_expr + arith_op + simple_expr >>
    unpack(OperatorAssignment)
)

simple_assignment = (
    variable + skip(mk_op('=')) + simple_expr >> unpack(SimpleAssignment))

function_call = (
    variable_list + skip(mk_op('=')) + function_name +
    skip(mk_op('(')) + simple_expr_list + skip(mk_op(')')) >>
    unpack(FunctionCall)
)

return_expr = (
    skip(mk_keyword('return')) + simple_expr_list >>
    unpack(Return)
)

return_fun = (
    skip(mk_keyword('returnf')) + function_name +
    lparen + simple_expr_list + rparen >>
    unpack(ReturnFun)
)

line_asm = (
    skip(mk_keyword("asm")) + function_name + many(number|variable|mk_op('[')|mk_op(']')|mk_op(',')) >>
    (lambda (n, a): Asm(n, a)))

line_expr = (
    newline |
    ((function_call | operator_assignment | simple_assignment | line_asm |
      return_expr | return_fun) + skip(newline))
)

expression = with_forward_decls(lambda:
    line_expr | simple_if | while_exp | function_def)

simple_if = (
    skip(mk_keyword('if')) + lparen + simple_expr + comp_op + simple_expr + rparen +
    skip(mk_op('{')) + skip(newline) +
        many(expression) +
    skip(mk_op('}')) + skip(mk_keyword('else')) + skip(mk_op('{')) + skip(newline) +
        many(expression) +
    skip(mk_op('}')) >>
    unpack(If)

)

while_exp = (
    mk_keyword('while')+ mk_op('(')+simple_expr+mk_op(')')+
    mk_op('{')+newline+
    many(expression) + mk_op('}') >>
    (lambda a: While(a[2],a[6]))
)

function_def = (
    skip(mk_keyword('def')) + function_name +
    lparen + variable_list + skip(mk_op(':')) +
            local_list + rparen + skip(mk_op('{')) + skip(newline) +
        many(expression) +
    skip(mk_op('}')) + skip(newline) >>
    unpack(FunctionDef)
)
skip_newlines = skip(maybe(many(newline)))
program = many(function_def + skip_newlines) + skip(endmarker) + skip(finished) >> Program

def test():
    print number.parse(tokenize('6'))
    print operator_assignment.parse(tokenize('a = 6 + b'))
    print simple_assignment.parse(tokenize('a = 6'))
    print variable_list.parse(tokenize('a, b, c'))
    print function_call.parse(tokenize('a = f(1,b,c)'))
    print return_expr.parse(tokenize('return a, 3, c'))
    print return_fun.parse(tokenize('returnf asd(a, 3, c)'))

    print function_def.parse(tokenize(
        'def f(a, b : c,d) {\n'
        '    a = 3\n'
        '}\n'
    ))

    print simple_if.parse(tokenize("""if (a > b) {
    a = 3
    } else {
    a = 4
    } """))
    print line_expr.parse(tokenize('return a, 3, c\n'))
    print line_expr.parse(tokenize('\n'))
    print '\n\n*****************'
    print line_expr.parse(tokenize('a = f(b,c)\n'))
    print line_expr.parse(tokenize('a = f\n'))

    #print function_def.parse(tokenize(file('f1.gcc').read()))
    #for t in tokenize(file('test.trup').read()):
    #    print unicode(t)

    print program.parse(tokenize(file('test.trup').read()))

    print line_asm.parse(tokenize("asm LDC 1"))
    print line_asm.parse(tokenize("asm LDC 2 1"))

    print '\n\n---'
    print expression.parse(tokenize("a = b\n"))
    print '---\n\n'

    print '\n\n---'
    print tokenize(file('test.trup').read())
    print program.parse(tokenize(file('test.trup').read()))
    print '---\n\n'

    print program.parse(tokenize(file('player.trup').read()))

    #print Assemble(program.parse(tokenize(file('test.trup').read())))
    outfile=file("a.out","w")
    print >> outfile, Assemble(program.parse(tokenize(file('player.trup').read())))


if __name__ == '__main__':
    import sys

    if len(sys.argv) == 1:
        test()
    elif len(sys.argv) == 3:
        trup = file(sys.argv[1]).read()
        gcc = str(Assemble(program.parse(tokenize(trup))))

        outfile = file(sys.argv[2], 'w')
        print >> outfile, gcc
    else:
        print "Usage:"
        print "  compile.py"
        print "  compile.py infile.trup outfile.gcc"


